@@ -32,7 +32,9 @@ class MappedSite(object):
         a Uniprot ID; 'UNIPROT_HTTP_NOT_FOUND' if the given Uniprot ID resulted
         in a 404 Not Found error from the Uniprot web service; or
         'UNIPROT_HTTP_OTHER' if it was any other type of Uniprot web service
-        error. If the error code is not None, the `orig_res` and `orig_pos`
+        error. Any other unexpected errors in getting the sequence are assigned
+        the 'UNIPROT_OTHER' code.
+        If the error code is not None, the `orig_res` and `orig_pos`
         fields will be set (based on the query arguments) but all other fields
         will be None.
     valid : bool
@@ -58,6 +60,9 @@ class MappedSite(object):
     gene_name : str
         The standard (HGNC) gene name of the protein that was mapped.
     """
+    attrs = ('up_id', 'error_code', 'valid', 'orig_res', 'orig_pos',
+             'mapped_id', 'mapped_res', 'mapped_pos', 'description',
+             'gene_name')
     def __init__(self, up_id, valid, orig_res, orig_pos,
                  error_code=None, mapped_id=None, mapped_res=None,
                  mapped_pos=None, description=None, gene_name=None):
@@ -76,42 +81,29 @@ class MappedSite(object):
         quote_args = lambda args: tuple([a if a in (None, True, False)
                                            else ("'%s'" % a) for a in args])
         return ("MappedSite(up_id=%s, error_code=%s, valid=%s, "
-                    "orig_res=%s, orig_pos=%s, mapped_id=%s, mapped_res=%s, "
-                    "mapped_pos=%s, description=%s, gene_name=%s)" %
+                "orig_res=%s, orig_pos=%s, mapped_id=%s, mapped_res=%s, "
+                "mapped_pos=%s, description=%s, gene_name=%s)" %
                 quote_args([self.up_id, self.error_code, self.valid,
                             self.orig_res, self.orig_pos, self.mapped_id,
                             self.mapped_res, self.mapped_pos, self.description,
                             self.gene_name]))
 
     def __eq__(self, other):
-        if (self.up_id == other.up_id and self.valid == other.valid and
-            self.error_code == other.error_code and
-            self.orig_res == other.orig_res and
-            self.orig_pos == other.orig_pos and
-            self.mapped_id == other.mapped_id and
-            self.mapped_res == other.mapped_res and
-            self.mapped_pos == other.mapped_pos and
-            self.description == other.description and
-            self.gene_name == other.gene_name):
-            return True
-        else:
-            return False
+        return hash(self) == hash(other)
 
     def __ne__(self, other):
         return not(self == other)
 
     def __hash__(self):
-        return hash((self.up_id, self.error_code, self.valid, self.orig_res,
-                     self.orig_pos, self.mapped_id, self.mapped_res,
-                     self.mapped_pos, self.description, self.gene_name))
+        return hash(tuple(self.to_list()))
 
     def to_json(self):
-        keys = ('up_id', 'error_code', 'valid', 'orig_res', 'orig_pos',
-                'mapped_id', 'mapped_res', 'mapped_pos', 'description',
-                'gene_name')
-        jd = {key: self.__dict__.get(key) for key in keys}
+        jd = {a: self.__dict__.get(a) for a in self.attrs}
         return jd
 
+    def to_list(self):
+        return [self.__getattribute__(a) for a in self.attrs]
+
     def not_invalid(self):
         """Return True if the original site is not known to be invalid.
 
@@ -236,7 +228,7 @@ class ProtMapper(object):
         mapped_sites = []
         for ix, (prot_id, prot_ns, residue, position) in enumerate(site_list):
             logger.info("Mapping site %d of %d, cache size %d" %
-                        (ix, len(site_list), len(self._cache)))
+                        (ix + 1, len(site_list), len(self._cache)))
             try:
                 ms = self.map_to_human_ref(prot_id, prot_ns, residue, position,
                                            **kwargs)
@@ -325,11 +317,16 @@ class ProtMapper(object):
         try:
             site_valid = uniprot_client.verify_location(up_id, residue,
                                                         position)
+            error_code = None
         except HTTPError as ex:
             if ex.response.status_code == 404:
                 error_code = 'UNIPROT_HTTP_NOT_FOUND'
             else:
                 error_code = 'UNIPROT_HTTP_OTHER'
+        except Exception as ex:
+            error_code = 'UNIPROT_OTHER'
+            logger.error(ex)
+        if error_code:
             # Set error_code; valid will set to None, not True/False
             mapped_site = MappedSite(up_id, None, residue, position,
                                      error_code=error_code)
@@ -360,9 +357,9 @@ class ProtMapper(object):
         # being shifted
         signal_peptide = uniprot_client.get_signal_peptide(up_id, False)
         # If there is valid signal peptide information from UniProt
-        if signal_peptide and signal_peptide[0] == 1 and \
-                signal_peptide[1] is not None:
-            offset_pos = str(int(position) + signal_peptide[1])
+        if signal_peptide and signal_peptide.begin == 1 and \
+                signal_peptide.end is not None:
+            offset_pos = str(int(position) + signal_peptide.end)
             # Check to see if the offset position is known to be phosphorylated
             mapped_site = self.get_psp_mapping(
                                 up_id, up_id, gene_name, residue, position,
@@ -484,11 +481,16 @@ class ProtMapper(object):
             site_valid = uniprot_client.verify_location(pspmapping.mapped_id,
                                       pspmapping.mapped_res,
                                       pspmapping.mapped_pos)
+            error_code = None
         except HTTPError as ex:
             if ex.response.status_code == 404:
                 error_code = 'UNIPROT_HTTP_NOT_FOUND'
             else:
                 error_code = 'UNIPROT_HTTP_OTHER'
+        except Exception as ex:
+            error_code = 'UNIPROT_OTHER'
+            logger.error(ex)
+        if error_code:
             # Set error_code; valid will set to None, not True/False
             mapped_site = MappedSite(orig_id, None, res, pos,
                                      error_code=error_code)
