@@ -1,13 +1,13 @@
-import os
 import re
 import csv
+import json
 import rdflib
 import logging
 import requests
-from xml.etree import ElementTree
 from functools import lru_cache
+from xml.etree import ElementTree
 from urllib.error import HTTPError
-from protmapper.resources import resource_manager
+from protmapper.resources import resource_manager, feature_from_json, Feature
 
 logger = logging.getLogger(__name__)
 
@@ -73,6 +73,27 @@ def query_protein(protein_id):
     return g
 
 
+def _strip_isoform(protein_id):
+    return protein_id.split('-')[0]
+
+
+def _split_isoform(protein_id):
+    parts = protein_id.split('-', maxsplit=1)
+    protein_id = parts[0]
+    isoform = None
+    if len(parts) == 2:
+        if re.match(r'\d+', parts[1]):
+            isoform = parts[1]
+    return protein_id, isoform
+
+
+def _reattach_isoform(pid, iso):
+    if iso is not None:
+        return '%s-%s' % (pid, iso)
+    else:
+        return pid
+
+
 def is_secondary(protein_id):
     """Return True if the UniProt ID corresponds to a secondary accession.
 
@@ -85,7 +106,7 @@ def is_secondary(protein_id):
     -------
     True if it is a secondary accessing entry, False otherwise.
     """
-    entry = um.uniprot_sec.get(protein_id)
+    entry = um.uniprot_sec.get(_strip_isoform(protein_id))
     if not entry:
         return False
     return True
@@ -103,11 +124,7 @@ def is_reviewed(protein_id):
     -------
     True if it is a reviewed entry, False otherwise.
     """
-    # If this is an isoform, check to see if the base ID is reviewed
-    # (isoform IDs are not in the ID list)
-    if '-' in protein_id:
-        protein_id = protein_id.split('-')[0]
-    return (protein_id in um.uniprot_reviewed)
+    return _strip_isoform(protein_id) in um.uniprot_reviewed
 
 
 def get_primary_id(protein_id):
@@ -126,20 +143,21 @@ def get_primary_id(protein_id):
         human one is returned. If there are no human primary IDs then the
         first primary found is returned.
     """
-    primaries = um.uniprot_sec.get(protein_id)
+    base_id, isoform = _split_isoform(protein_id)
+    primaries = um.uniprot_sec.get(base_id)
     if primaries:
         if len(primaries) > 1:
-            logger.debug('More than 1 primary ID for %s.' % protein_id)
+            logger.debug('More than 1 primary ID for %s.' % base_id)
             for primary in primaries:
                 # Often secondary IDs were broken into multiple primary IDs
                 # for different organisms. In this case we return the human
                 # one if it exists.
                 if is_human(primary):
-                    return primary
+                    return _reattach_isoform(primary, isoform)
         # If we haven't returned anything then we just return the
         # first primary id
-        return primaries[0]
-    # If there is not secondary entry the we assume this is a primary entry
+        return _reattach_isoform(primaries[0], isoform)
+    # If there is no secondary entry then we assume this is a primary entry
     return protein_id
 
 
@@ -192,7 +210,7 @@ def get_mnemonic(protein_id, web_fallback=False):
     mnemonic : str
         The UniProt mnemonic corresponding to the given Uniprot ID.
     """
-    protein_id = protein_id.split('-', maxsplit=1)[0]
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     try:
         mnemonic = um.uniprot_mnemonic[protein_id]
         return mnemonic
@@ -253,7 +271,7 @@ def get_gene_name(protein_id, web_fallback=True):
     gene_name : str
         The gene name corresponding to the given Uniprot ID.
     """
-    protein_id = get_primary_id(protein_id)
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     try:
         gene_name = um.uniprot_gene_name[protein_id]
         # We only get here if the protein_id was in the dict
@@ -300,6 +318,7 @@ def get_gene_synonyms(protein_id):
     synonyms : list[str]
         The list of synonyms of the gene corresponding to the protein
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     g = query_protein(protein_id)
     if g is None:
         return None
@@ -332,6 +351,7 @@ def get_protein_synonyms(protein_id):
     synonyms : list[str]
         The list of synonyms of the protein
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     g = query_protein(protein_id)
     if g is None:
         return None
@@ -360,6 +380,7 @@ def get_synonyms(protein_id):
     synonyms : list[str]
         The list of synonyms of the protein and its associated gene.
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     ret = []
     gene_syms = get_gene_synonyms(protein_id)
     if gene_syms:
@@ -372,17 +393,12 @@ def get_synonyms(protein_id):
 
 @lru_cache(maxsize=10000)
 def get_sequence(protein_id):
-    # Get the primary ID
-    try:
-        prim_ids = um.uniprot_sec[protein_id]
-        protein_id = prim_ids[0]
-    except KeyError:
-        pass
+    base, iso = _split_isoform(get_primary_id(protein_id))
     # Try to get the sequence from the downloaded sequence files
-    if '-' in protein_id:
-        base, iso = protein_id.split('-')
-        if iso == '1':
-            protein_id = base
+    if iso == '1':
+        protein_id = base
+    else:
+        protein_id = _reattach_isoform(base, iso)
     seq = um.uniprot_sequences.get(protein_id)
     if seq is None:
         url = uniprot_url + '%s.fasta' % protein_id
@@ -395,6 +411,7 @@ def get_sequence(protein_id):
 
 
 def get_modifications(protein_id):
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     g = query_protein(protein_id)
     if g is None:
         return None
@@ -498,6 +515,7 @@ def verify_modification(protein_id, residue, location=None):
 
 
 def _is_organism(protein_id, organism_suffix):
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     mnemonic = get_mnemonic(protein_id)
     if mnemonic is None:
         return False
@@ -518,6 +536,7 @@ def is_human(protein_id):
     -------
     True if the protein_id corresponds to a human protein, otherwise False.
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     return _is_organism(protein_id, 'HUMAN')
 
 
@@ -533,6 +552,7 @@ def is_mouse(protein_id):
     -------
     True if the protein_id corresponds to a mouse protein, otherwise False.
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     return _is_organism(protein_id, 'MOUSE')
 
 
@@ -548,6 +568,7 @@ def is_rat(protein_id):
     -------
     True if the protein_id corresponds to a rat protein, otherwise False.
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     return _is_organism(protein_id, 'RAT')
 
 
@@ -564,6 +585,7 @@ def get_hgnc_id(protein_id):
     hgnc_id : str
         HGNC ID of the human protein
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     return um.uniprot_hgnc.get(protein_id)
 
 
@@ -580,6 +602,7 @@ def get_mgi_id(protein_id):
     mgi_id : str
         MGI ID of the mouse protein
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     return um.uniprot_mgi.get(protein_id)
 
 
@@ -596,6 +619,7 @@ def get_rgd_id(protein_id):
     rgd_id : str
         RGD ID of the rat protein
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     return um.uniprot_rgd.get(protein_id)
 
 
@@ -644,6 +668,7 @@ def get_mouse_id(human_protein_id):
     mouse_protein_id : str
         The UniProt ID of a mouse protein orthologous to the given human protein
     """
+    human_protein_id = get_primary_id(_strip_isoform(human_protein_id))
     return um.uniprot_human_mouse.get(human_protein_id)
 
 
@@ -660,6 +685,7 @@ def get_rat_id(human_protein_id):
     rat_protein_id : str
         The UniProt ID of a rat protein orthologous to the given human protein
     """
+    human_protein_id = get_primary_id(_strip_isoform(human_protein_id))
     return um.uniprot_human_rat.get(human_protein_id)
 
 
@@ -676,6 +702,7 @@ def get_length(protein_id):
     length : int
         The length of the protein in amino acids.
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     return um.uniprot_length.get(protein_id)
 
 
@@ -697,11 +724,7 @@ def query_protein_xml(protein_id):
         An ElementTree representation of the XML entry for the
         protein.
     """
-    try:
-        prim_ids = um.uniprot_sec[protein_id]
-        protein_id = prim_ids[0]
-    except KeyError:
-        pass
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     url = uniprot_url + protein_id + '.xml'
     try:
         ret = requests.get(url)
@@ -734,6 +757,40 @@ def get_function(protein_id):
     return function.text
 
 
+def get_features(protein_id):
+    """Return a list of features (chains, peptides) for a given protein.
+
+    Parameters
+    ----------
+    protein_id : str
+        The UniProt ID of the protein whose features are to be returned.
+
+    Returns
+    -------
+    list of Feature
+        A list of Feature named tuples representing each Feature.
+    """
+    return um.features.get(protein_id, [])
+
+
+def get_chains(protein_id):
+    """Return the list of cleaved chains for the given protein.
+
+    Parameters
+    ----------
+    protein_id : str
+        The UniProt ID of the protein whose cleaved chains are to be returned.
+
+    Returns
+    -------
+    list of Feature
+        A list of Feature named tuples representing each chain.
+    """
+    features = get_features(protein_id)
+    chains = [f for f in features if f.type == 'CHAIN']
+    return chains
+
+
 def get_signal_peptide(protein_id, web_fallback=True):
     """Return the position of a signal peptide for the given protein.
 
@@ -748,17 +805,23 @@ def get_signal_peptide(protein_id, web_fallback=True):
 
     Returns
     -------
-    tuple of int
-        The beginning and end position of the signal peptide as a tuple
-        of integers.
+    Feature
+        A Feature named tuple representing the signal peptide.
     """
+    protein_id = get_primary_id(_strip_isoform(protein_id))
     # Note, we use False here to differentiate from None
-    entry = um.signal_peptide.get(protein_id, False)
-    if entry is not False or not web_fallback:
-        return entry
+    if not web_fallback and protein_id not in um.features:
+        return None
+    elif protein_id in um.features:
+        sp = [f for f in get_features(protein_id) if f.type == 'SIGNAL']
+        if sp:
+            return sp[0]
+        elif not web_fallback:
+            return False
+
     et = query_protein_xml(protein_id)
     if et is None:
-        return None, None
+        return None
     location = et.find(
         'up:entry/up:feature[@type="signal peptide"]/up:location',
         namespaces=xml_ns)
@@ -775,7 +838,25 @@ def get_signal_peptide(protein_id, web_fallback=True):
             end_pos = end.attrib.get('position')
             if end_pos is not None:
                 end_pos = int(end_pos)
-    return begin_pos, end_pos
+    if begin_pos is not None and end_pos is not None:
+        return Feature('SIGNAL', begin_pos, end_pos, None, None)
+    return None
+
+
+def get_feature_by_id(feature_id):
+    """Return a Feature based on its unique feature ID.
+
+    Parameters
+    ----------
+    feature_id : str
+        A Feature ID, of the form PRO_*.
+
+    Returns
+    -------
+    Feature or None
+        A Feature with the given ID.
+    """
+    return um.features_by_id.get(feature_id)
 
 
 def get_ids_from_refseq(refseq_id, reviewed_only=False):
@@ -818,7 +899,7 @@ class UniprotMapper(object):
          self._uniprot_mnemonic_reverse, self._uniprot_mgi,
          self._uniprot_rgd, self._uniprot_mgi_reverse,
          self._uniprot_rgd_reverse, self._uniprot_length,
-         self._uniprot_reviewed, self._uniprot_signal_peptide) = maps
+         self._uniprot_reviewed, self._features, self._features_by_id) = maps
 
         self._uniprot_sec = _build_uniprot_sec()
 
@@ -846,7 +927,7 @@ class UniprotMapper(object):
 
     @property
     def uniprot_hgnc(self):
-        if not self.initialized:
+        if not self.initialized_hgnc:
             self.initialize_hgnc()
         return self._uniprot_hgnc
 
@@ -929,10 +1010,16 @@ class UniprotMapper(object):
         return self._refseq_uniprot
 
     @property
-    def signal_peptide(self):
+    def features(self):
         if not self.initialized:
             self.initialize()
-        return self._uniprot_signal_peptide
+        return self._features
+
+    @property
+    def features_by_id(self):
+        if not self.initialized:
+            self.initialize()
+        return self._features_by_id
 
 
 um = UniprotMapper()
@@ -940,6 +1027,7 @@ um = UniprotMapper()
 
 def _build_uniprot_entries():
     up_entries_file = resource_manager.get_create_resource_file('up')
+    sc_entries_file = resource_manager.get_create_resource_file('up_sars_cov2')
     uniprot_gene_name = {}
     uniprot_mnemonic = {}
     uniprot_mnemonic_reverse = {}
@@ -948,43 +1036,46 @@ def _build_uniprot_entries():
     uniprot_mgi_reverse = {}
     uniprot_rgd_reverse = {}
     uniprot_length = {}
-    uniprot_signal_peptide = {}
+    uniprot_features = {}
     uniprot_reviewed = set()
-    with open(up_entries_file, 'r') as fh:
-        csv_rows = csv.reader(fh, delimiter='\t')
-        # Skip the header row
-        next(csv_rows)
-        for row in csv_rows:
-            up_id, gene_name, up_mnemonic, rgd, mgi, length, reviewed, \
-                signal_peptide = row
-            # Store the entry in the reviewed set
-            if reviewed == 'reviewed':
-                uniprot_reviewed.add(up_id)
-            uniprot_gene_name[up_id] = gene_name
-            uniprot_mnemonic[up_id] = up_mnemonic
-            uniprot_mnemonic_reverse[up_mnemonic] = up_id
-            uniprot_length[up_id] = int(length)
-            if mgi:
-                mgi_ids = mgi.split(';')
-                if mgi_ids:
-                    uniprot_mgi[up_id] = mgi_ids[0]
-                    uniprot_mgi_reverse[mgi_ids[0]] = up_id
-            if rgd:
-                rgd_ids = rgd.split(';')
-                if rgd_ids:
-                    uniprot_rgd[up_id] = rgd_ids[0]
-                    uniprot_rgd_reverse[rgd_ids[0]] = up_id
-            uniprot_signal_peptide[up_id] = (None, None)
-            if signal_peptide:
-                match = re.match(r'SIGNAL (\d+) (\d+) ', signal_peptide)
-                if match:
-                    beg_pos, end_pos = match.groups()
-                    uniprot_signal_peptide[up_id] = \
-                        (int(beg_pos), int(end_pos))
+    files = [up_entries_file, sc_entries_file]
+    for file in files:
+        with open(file, 'r') as fh:
+            csv_rows = csv.reader(fh, delimiter='\t')
+            # Skip the header row
+            next(csv_rows)
+            for row in csv_rows:
+                up_id, gene_name, up_mnemonic, rgd, mgi, length, reviewed, \
+                    features_json = row
+                # Store the entry in the reviewed set
+                if reviewed == 'reviewed':
+                    uniprot_reviewed.add(up_id)
+                uniprot_gene_name[up_id] = gene_name
+                uniprot_mnemonic[up_id] = up_mnemonic
+                uniprot_mnemonic_reverse[up_mnemonic] = up_id
+                uniprot_length[up_id] = int(length)
+                if mgi:
+                    mgi_ids = mgi.split(';')
+                    if mgi_ids:
+                        uniprot_mgi[up_id] = mgi_ids[0]
+                        uniprot_mgi_reverse[mgi_ids[0]] = up_id
+                if rgd:
+                    rgd_ids = rgd.split(';')
+                    if rgd_ids:
+                        uniprot_rgd[up_id] = rgd_ids[0]
+                        uniprot_rgd_reverse[rgd_ids[0]] = up_id
+                uniprot_features[up_id] = [feature_from_json(feat) for
+                                           feat in json.loads(features_json)]
+
+    # Build a dict of features by feature ID
+    features_by_id = {}
+    for up_id, feats in uniprot_features.items():
+        for feat in feats:
+            features_by_id[feat.id] = feat
 
     return (uniprot_gene_name, uniprot_mnemonic, uniprot_mnemonic_reverse,
             uniprot_mgi, uniprot_rgd, uniprot_mgi_reverse, uniprot_rgd_reverse,
-            uniprot_length, uniprot_reviewed, uniprot_signal_peptide)
+            uniprot_length, uniprot_reviewed, uniprot_features, features_by_id)
 
 
 def _build_human_mouse_rat():
